﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Proteus;

namespace Spark.Installer
{
    public class InstallerVariable
    {
        public const string ProgramFiles = "[PROGRAM_FILES]";
        public const string ProgramFilesX86 = "[PROGRAM_FILES_X86]";
        public const string InstallDir = "[INSTALLDIR]";
        public const string Desktop = "[DESKTOP]";
        public const string SystemRoot = "[SYSTEMROOT]";
        public const string UserStartup = "[USER_STARTUP]";  // startup folder
    }
    public class InstallOptionValue
    {
        public const string True = "True";
        public const string False = "True";
    }
    //key value pairs
    public class InstallOption
    {
        //Constants.
        public const string DefaultDirectory = "DefaultDirectory";
        //public const string CreateDesktopShortcut = "CreateDesktopShortcut";
        //public const string CreateStartMenuFolder = "CreateStartMenuFolder";
        //Program Info
        public const string DisplayName = "DisplayName";
        public const string InstallLocation = "InstallLocation";
        public const string Publisher = "Publisher";
        public const string DisplayIcon = "DisplayIcon";
        public const string ApplicationVersion = "ApplicationVersion";
        public const string DisplayVersion = "DisplayVersion";
        public const string URLInfoAbout = "URLInfoAbout";
        public const string Contact = "Contact";
        public const string UninstallFolderRoot_Uninstaller_Only = "UninstallFolderRoot_Uninstaller_Only"; //**DO NOT PACKAGE THIS WITH THE INSTALLER. This determines whether we are running from installer
        public const string AppGuid = "AppGuid";
        public const string ExecuteSync = "ExecuteSync";
        public const string DeleteFile = "DeleteFile";
        public const string Shortcut = "Shortcut";

        public string Key;
        public string Value;
        public InstallerSectionType SectionType;

        public InstallOption(string key, string value, InstallerSectionType sectype)
        {
            Key = key;
            Value = value;
            SectionType = sectype;
        }
        public byte[] Serialize()
        {
            byte[] ret = new byte[0];
            ret = BufferUtils.Combine(ret, StringUtils.SerializeUTF8Path(Key));
            ret = BufferUtils.Combine(ret, StringUtils.SerializeUTF8Path(Value));
            ret = BufferUtils.Combine(ret, BitConverter.GetBytes((Int64)SectionType));
            return ret;
        }
        public static InstallOption Deserialize(InstallerBinary binary)
        {
            string key = binary.ReadUTF8Path();
            string value = binary.ReadUTF8Path();
            Int64 sectionType = binary.ReadInt64();
            return new InstallOption(key, value, (InstallerSectionType)sectionType);
        }
        public ShortcutOption TryParseToShortcutOption(InstallManager mgr, ProgramInfo pi)
        {
            ShortcutOption sc = new ShortcutOption(Key, Value, SectionType);

            string[] values = Value.Split(InstallerConfig.OptionArgumentDelimiter);
            sc.ShortcutLocation = (ShortcutLocation)Enum.Parse(typeof(ShortcutLocation), StringUtils.Dequote(values[0]));

            string strRequired = StringUtils.Dequote(values[1]).ToLower();
            if (strRequired.Equals("required"))
                sc.Required = true;
            else if (strRequired.Equals("optional"))
                sc.Required = false;
            else
                throw new Exception("Shortcut requirement is invalid:" + strRequired);

            sc.TargetPath = StringUtils.Dequote(values[2]);
            sc.ShortcutName = StringUtils.Dequote(values[3]);
            sc.Arguments = StringUtils.Dequote(values[4]);

            sc.ParseDetails(mgr, pi);

            return sc;
        }
    }

    public class ShortcutOption : InstallOption
    {
        public bool Required;
        public ShortcutLocation ShortcutLocation; // 
        public string TargetPath; // where we put the shortcut
        public string Arguments; //arguemtns to app
        public string ShortcutName; //name of shortcut
        public string ShortcutPath;
        public string IconPath;
        public bool Enabled;
        public string WorkingDirectory;
        public ShortcutOption(string key, string value, InstallerSectionType secType) : base(key,value,secType)
        {
            Enabled = true; // default.  If these are optional then user can disable them
        }
        public void ParseDetails(InstallManager mgr, ProgramInfo objProgramInfo)
        {
            //Note: we allow programinfo to be null because we have to preload the options in the main menu and we won't have the programinfo constructed yet.
            // Get Details
            string folder = string.Empty;
            string cpyPath = string.Empty;
            // Get path to shortcut.
            if (ShortcutLocation == ShortcutLocation.Desktop)
            {
                folder = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);
                cpyPath = folder;
            }
            else if (ShortcutLocation == ShortcutLocation.StartMenu)
            {
                folder = Environment.GetFolderPath(Environment.SpecialFolder.Programs);
                cpyPath = System.IO.Path.Combine(folder, objProgramInfo.DisplayName);
                cpyPath = mgr.ReplaceValueVariables(cpyPath);
            }
            else if (ShortcutLocation == ShortcutLocation.Startup)
            {
                folder = Environment.GetFolderPath(Environment.SpecialFolder.Startup);
                //Don't specify folders in startup or windows will open the folder.
                cpyPath = folder;
            }
            else
                throw new Exception("Failed to identify installer option shortcut location: " + ShortcutLocation);

            // To some dirty.
            if (!System.IO.Directory.Exists(cpyPath))
                System.IO.Directory.CreateDirectory(cpyPath);

            string shortcutName = ShortcutName.Replace(".lnk", "");//remove superfuluous .lnk
            string shortcutPath = System.IO.Path.Combine(cpyPath, ShortcutName + ".lnk");
            string shortcutTarget = TargetPath;
            string shortcutArgs = Arguments;
            string iconPath = objProgramInfo.DisplayIcon;

            ShortcutName = mgr.ReplaceValueVariables(shortcutName);
            ShortcutPath = mgr.ReplaceValueVariables(shortcutPath);
            TargetPath   = mgr.ReplaceValueVariables(shortcutTarget);
            Arguments    = mgr.ReplaceValueVariables(shortcutArgs);
            IconPath     = mgr.ReplaceValueVariables(iconPath);

            string dirname = System.IO.Path.GetDirectoryName(TargetPath);
            WorkingDirectory = dirname;
        }
    }
    public class InstallOptions
    {
        private ProgramInfo _objProgramInfo;
        private List<ShortcutOption> _lstShortcutOptions = new List<ShortcutOption>();
        private List<InstallOption> _lstInstallOptions = new List<InstallOption>();

        public ProgramInfo GetProgramInfo()
        {
            return _objProgramInfo;
        }
        public List<ShortcutOption> GetShortcutOptions()
        {
            return _lstShortcutOptions;
        }
        public List<InstallOption> GetAllOptionsForSection(InstallerSectionType section)
        {
            List<InstallOption> opts = _lstInstallOptions.Where(x => x.SectionType.Equals(section)).ToList();
            return opts;
        }
        public InstallOption GetUniqueOption(string key)
        {
            //Get a SINGLE UNIQUE option - there cannot be more than one of this option type.
            List<InstallOption> opts = GetOptions(key);
            if (opts.Count == 0)
                return null;
            if (opts.Count > 1)
                //throw here. t his is so we don't thwart the old implementation which allowed only one parameter per option.
                throw new Exception("Tried to get multi-option parameter.");

            return opts[0];
        }
        public List<InstallOption> GetOptions(string key)
        {
            List<InstallOption> opt = _lstInstallOptions.Where(x => x.Key.Equals(key)).ToList();
            return opt;
        }
        public string GetOptionValueOrDefault(string key)
        {
            List<InstallOption> opts = GetOptions(key);
            if (opts.Count > 1)
                //throw here. t his is so we don't thwart the old implementation which allowed only one parameter per option.
                throw new Exception("Tried to get multi-option parameter.");

            if (opts.Count==0)
                return string.Empty;

            return opts[0].Value;
        }
        
        public void AddOrReplaceOption(InstallOption opt)
        {
            InstallOption found = GetUniqueOption(opt.Key);
            if (found != null)
            {
                found.Value = opt.Value;
            }
            else
            {
                _lstInstallOptions.Add(opt);
            }
        }
        public List<InstallOption> Options
        {
            get
            {
                return _lstInstallOptions;
            }
            set
            {
                _lstInstallOptions = value;
            }
        }
        public void Deserialize(InstallerBinary objBinary, InstallManager mgr)
        {
            Int64 count = objBinary.ReadInt64();
            for (Int64 iOption = 0; iOption < count; iOption++)
            {
                InstallOption opt = InstallOption.Deserialize(objBinary);
                _lstInstallOptions.Add(opt);
            }

            AfterDeserialize(mgr);
        }
        public byte[] Serialize()
        {
            byte[] ret = new byte[0];
            byte[] tmp;

            Int64 numEntries = Options.Count;
            ret = BufferUtils.Combine(ret, BitConverter.GetBytes(numEntries));
            foreach (InstallOption ent in Options)
            {
                tmp = ent.Serialize();
                ret = BufferUtils.Combine(ret, tmp);
            }

            return ret;
        }
        public void Reload(InstallManager mgr) {
            AfterDeserialize(mgr);
        }

        private void AfterDeserialize(InstallManager mgr)
        {
            _objProgramInfo = new ProgramInfo(this);
            _lstShortcutOptions = ParseShortcutOptions(mgr);
        }
        private List<ShortcutOption> ParseShortcutOptions(InstallManager mgr)
        {
            List<ShortcutOption> ret = new List<ShortcutOption>();
            List<InstallOption> options = GetOptions(InstallOption.Shortcut);

            foreach (InstallOption opt in options)
            {
                ShortcutOption sc = opt.TryParseToShortcutOption(mgr, _objProgramInfo);
                ret.Add(sc);
            }

            return ret;
        }
    }
}
