﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Proteus
{

    public class ConsoleProcess
    {
        #region Props Private

        private System.ComponentModel.BackgroundWorker bw;
        private System.Diagnostics.Process ShellProcess { get; set; }
        private string _strOutput;
        private Object _objLockObject = new Object();
        private Object _objOutputLockObject = new Object();
        private Object _objMessageLockObject = new Object();
        private ConsoleProcessCommandState _eCommandState;
        private Queue<ConsoleProcessCommand> _commandsToThread { get; set; }
        private Queue<ConsoleProcessCommand> _commandsToHost { get; set; }
        private int _intProcessorIndex;
        private bool _blnSynchronous;

        #endregion

        #region Props Public
        public bool Success;
        private bool _blnHasStarted = false;
        public bool HasStarted
        {
            get
            {
                return _blnHasStarted;
            }
        }
        public bool IsRunning
        {
            get
            {
                return (CommandState == ConsoleProcessCommandState.Running);
            }
        }
        public string CommandOutput
        {
            get
            {
                lock (_objOutputLockObject)
                {
                    return _strOutput;
                }
            }
            private set
            {
                lock (_objOutputLockObject)
                {
                    _strOutput = value;
                }
            }
        }
        public ConsoleProcessCommandState CommandState
        {
            get
            {
                lock (_objLockObject)
                {
                    return _eCommandState;
                }

            }
            private set
            {
                lock (_objLockObject)
                {
                    _eCommandState = value;
                }
            }
        }
        protected System.ComponentModel.BackgroundWorker BackgroundWorker
        {
            get
            {
                lock (_objLockObject)
                {
                    return bw;
                }

            }
            private set
            {
                lock (_objLockObject)
                {
                    bw = value;
                }
            }
        }
        private bool _blnCreateWindow;
        #endregion

        public ConsoleProcess(bool synchronous = false, bool showConsoleWindow = true)
        {
            _commandsToThread = new Queue<ConsoleProcessCommand>();
            _commandsToHost = new Queue<ConsoleProcessCommand>();
            _blnSynchronous = synchronous;
            _blnCreateWindow = showConsoleWindow;
        }

        #region Public: Methods
       
        public void BeginAsync(int iProcessorIndex = -1)
        {
            _intProcessorIndex = iProcessorIndex;

            bw = new System.ComponentModel.BackgroundWorker();
            bw.DoWork += WaitForCommands;
            //bw.RunWorkerCompleted += WorkerCompleted;
            bw.WorkerSupportsCancellation = true;
            bw.RunWorkerAsync();
        }
        public void Execute(string cmdText, ulong timeoutMs = 0, System.Threading.CancellationToken token = default(System.Threading.CancellationToken))
        {
            //**Note: if synchronous and the process doesn't exit, the command will block.
            SendMessage(new ConsoleProcessCommand(cmdText, ConsoleProcessCommandType.ExecuteCommandText, _blnSynchronous, timeoutMs, token));
        }
        public void Exit(bool waitForThreadToDie, ulong timeout = 4000)
        {
            ulong tA = Win32Utils.GetTickCount64();
            //Wait for idle.
            while(CommandState == ConsoleProcessCommandState.Running && _commandsToThread.Count>0)
            {
                if (timeout>0 && (Win32Utils.GetTickCount64() - tA) > timeout)
                {
                    Globals.Logger.LogWarn("Timeout exceeded for concoel porocess.");
                    break;
                }
                System.Windows.Forms.Application.DoEvents();
                CheckUserCancellation();
            }
            SendMessage(new ConsoleProcessCommand("yes", ConsoleProcessCommandType.Exit));
            while (CommandState != ConsoleProcessCommandState.Exited)
            {
                System.Windows.Forms.Application.DoEvents();
                CheckUserCancellation();
            }
            BackgroundWorker.CancelAsync();

        }
        public void SendMessage(ConsoleProcessCommand cmd)
        {
            lock (_objMessageLockObject)
            {
                _commandsToThread.Enqueue(cmd);
            }
        }
        public ConsoleProcessCommand GetMessageFromThread()
        {
            ConsoleProcessCommand cmd = null;

            lock (_objMessageLockObject)
            {
                if (_commandsToHost.Count > 0)
                    cmd = _commandsToHost.Dequeue();
            }
            return cmd;
        }
        
        #endregion

        #region Protected: Methods

        protected virtual void ProcessCommand(ConsoleProcessCommand cmd)
        {
            //bool blnReturn = true;
            switch (cmd.CommandType)
            {
                case ConsoleProcessCommandType.ExecuteCommandText:
                    lock (_objLockObject)
                    {
                        CommandState = ConsoleProcessCommandState.Running;
                        _blnHasStarted = true;
                        if (_blnSynchronous)
                        {
                            ExecuteCommandSync(cmd.CommandText, cmd.TimeoutInMilliseconds);
                        }
                        else
                        {
                            ShellProcess.StandardInput.WriteLine("rem #BEGIN#");
                            ShellProcess.StandardInput.WriteLine(cmd.CommandText);
                            ShellProcess.StandardInput.WriteLine("rem #COMPLETE#");
                        }
                    }
                    break;
                case ConsoleProcessCommandType.Exit:
                    lock (_objLockObject)
                    {
                        if (_blnSynchronous)
                        {
                            //Manually set exit
                            CommandState = ConsoleProcessCommandState.Exited;
                        }
                        else
                        {
                            // write exit, and wait for process to read it before quitting.
                            ShellProcess.StandardInput.WriteLine("rem #EXIT#");
                            ShellProcess.StandardInput.BaseStream.Flush();
                        }

                        //if (CommandState == CommandState.Running)
                        //    WaitForLastExecutedCommand();

                       // blnReturn = false;
                    }
                    break;
                default:
                    throw new NotImplementedException();
            }

           // return blnReturn;
        }
        
        #endregion

        #region Private: Methods
        private void FinishExecuting(string output, bool success)
        {
            //**Set output in our proc
            CommandOutput = output;
            CommandOutput += "##end##\n";

            Success = success;
            CommandState = ConsoleProcessCommandState.Idle;
        }
        private void ExecuteCommandSync(string text, ulong timeoutInMilliseconds)
        {
            //**Get all command data then call WaitForExit.  This will block the thread and prevent hte
            // cancellation token from executing.

            string processName;
            string processArgs;
            StringBuilder strOutput = new StringBuilder();
            StringBuilder strError = new StringBuilder();
            bool blnSuccess = true;
            bool blnWaitForExitResult = true; //must default to true
            try
            {
                text = text.Trim();
                processName = text.Substring(0, text.IndexOf(" "));
                processArgs = text.Substring(text.IndexOf(" "));
            }
            catch (Exception ex)
            {
                Globals.Logger.LogError(ex.ToString());
                FinishExecuting("Failed to get process name from string.", false);
                return;
            }

            //Note: the using is probably not necessary.  I did this to make sure the process wrote all the output but it should do it anyway.
            using (System.Diagnostics.Process tempProcess = new System.Diagnostics.Process())
            {
                System.Diagnostics.ProcessStartInfo inf = new System.Diagnostics.ProcessStartInfo(processName, processArgs);

                inf.WindowStyle = System.Diagnostics.ProcessWindowStyle.Normal;
                inf.CreateNoWindow = !_blnCreateWindow;
                inf.UseShellExecute = false;
                //inf.RedirectStandardInput = true;
                inf.RedirectStandardOutput = true;
                inf.RedirectStandardError = true;
                inf.ErrorDialog = false;

                tempProcess.OutputDataReceived += (sender, args) =>
                {
                    if (args.Data != null)
                    {
                        strOutput.Append(args.Data + Environment.NewLine);
                    }
                };
                tempProcess.ErrorDataReceived += (sender, args) =>
                {
                    if (args.Data != null)
                    {
                        strError.Append(args.Data + Environment.NewLine);
                    }
                };

                tempProcess.StartInfo = inf;

                //**Main CL Program execute
                try
                {
                    // STart will fail if process not found.
                    tempProcess.Start();
                    tempProcess.BeginErrorReadLine();
                    tempProcess.BeginOutputReadLine();

                    // Processes cannot exit until all their output is written.
                    // if the output of hte process fills the output buffer the buffer
                    // must be read.
                    if(timeoutInMilliseconds>0)
                        blnWaitForExitResult = tempProcess.WaitForExit((int)timeoutInMilliseconds);
                    else
                        tempProcess.WaitForExit();
                    if (blnWaitForExitResult)
                    {
                        tempProcess.CancelErrorRead(); // Is this necessary?
                        tempProcess.CancelOutputRead();
                    }
                    else
                    {
                        tempProcess.Kill();
                        blnSuccess = false;
                    }

                }
                catch (Exception ex)
                {
                    Globals.Logger.LogError(ex.ToString());
                    strOutput.Append(ex.ToString());
                }

                tempProcess.Close();
                tempProcess.Dispose();

            }
            FinishExecuting(strOutput.ToString() + strError.ToString(), blnSuccess);
        } 

        private ConsoleProcessCommand ThreadGetMessage()
        {
            ConsoleProcessCommand cmd = null;

            //System.Threading.Monitor.TryEnter(_objLockObject, 10);
            lock (_objMessageLockObject)
            {
                if (_commandsToThread.Count > 0)
                    cmd = _commandsToThread.Dequeue();
            }
            return cmd;
        }
        private void ThreadSendMessage(ConsoleProcessCommand cmd)
        {
            lock (_objMessageLockObject)
            {
                _commandsToHost.Enqueue(cmd);
            }
        }
        private void InvokeConsole()
        {
            if (_blnSynchronous)
            {

            }
            else
            {
                ShellProcess = CreateShellProcess();
            
                // Must read async see the vc documentation on ReadToEnd().. it will deadlock
                ShellProcess.OutputDataReceived += new System.Diagnostics.DataReceivedEventHandler(DataReceivedEventHandler);
                ShellProcess.BeginOutputReadLine();

            }
        }
        private System.Diagnostics.Process CreateShellProcess()
        {
            System.Diagnostics.ProcessStartInfo inf = new System.Diagnostics.ProcessStartInfo("cmd.exe", "/K ");
            inf.WindowStyle = System.Diagnostics.ProcessWindowStyle.Normal;
            //inf.CreateNoWindow = true;
            inf.UseShellExecute = false;
            inf.RedirectStandardInput = true;
            inf.RedirectStandardOutput = true;
            
            System.Diagnostics.Process ret = System.Diagnostics.Process.Start(inf);

            return ret;
        }
        private void DataReceivedEventHandler(object sender, System.Diagnostics.DataReceivedEventArgs e)
        {
            if (!String.IsNullOrEmpty(e.Data))
            {
                if (e.Data.Contains("#EXIT#"))
                {
                    //We got all data
                    CommandState = ConsoleProcessCommandState.Exited;
                    return;
                }
                if (e.Data.Contains("#COMPLETE#"))
                {
                    //We got all data
                    CommandState = ConsoleProcessCommandState.Idle;
                    return;
                }
                CommandOutput += e.Data;
                CommandOutput += "\n";
                Console.Write("[" + System.Threading.Thread.CurrentThread.ManagedThreadId + "] ");
                Console.ForegroundColor = ConsoleColor.Cyan;
                Console.WriteLine(e.Data);
                Console.ForegroundColor = ConsoleColor.White;
            }
        }
        private System.Threading.CancellationToken _objCancellationToken;
        private void WaitForCommands(object sender, System.ComponentModel.DoWorkEventArgs e)
        {
            CommandState = ConsoleProcessCommandState.Idle;

            InvokeConsole();
            while (true)
            {
                ConsoleProcessCommand rcvCmd = ThreadGetMessage();
                if (rcvCmd != null &&
                    CommandState != ConsoleProcessCommandState.Exited)
                {
                    ProcessCommand(rcvCmd);

                    if (rcvCmd.Token != System.Threading.CancellationToken.None)
                        _objCancellationToken = rcvCmd.Token;
                }
                else if (CommandState == ConsoleProcessCommandState.Exited)
                    break;
                else
                {
                    //Wait a bit to process messages
                    System.Windows.Forms.Application.DoEvents();
                }
                CheckUserCancellation();
            }

        }
        private void CheckUserCancellation()
        {
            //The given Optional token was set to cancel - kill this guy
            //throw and exit.
            if (_objCancellationToken != System.Threading.CancellationToken.None)
            {
                if (_objCancellationToken.IsCancellationRequested)
                {
                    _objCancellationToken.ThrowIfCancellationRequested();
                }
            }
        }
        //private void WaitForLastExecutedCommand(ulong timeout = (10*60*1000))
        //{
        //    ulong a = Win32Utils.GetTickCount64();
        //    while (CommandState != ConsoleProcessCommandState.Idle &&
        //        CommandState != ConsoleProcessCommandState.Exited)
        //    {
        //        ulong b = Win32Utils.GetTickCount64();
        //        if (b - a > timeout)
        //        {
        //            CommandOutput += " [ERROR] Process exceeded max timout limit of " + timeout.ToString() + "ms";
        //            return;
        //        }
        //        System.Threading.Thread.Sleep(100);
        //    }
        //    //Note we set the processor state ADFTER we have sent the file.
        //}
       
        #endregion

    }
}
